import os
from pynput import keyboard
import socket
import threading


# Startup file statics
STARTUP_FILE = os.getenv('APPDATA') + '\\Microsoft\\Windows\\Start Menu\\Programs\\Startup\\multi-clipboard.vbs'
STARTUP_COMMAND = 'python ' + os.path.dirname(os.path.realpath(__file__)) + '\\multi_clipboard.py'

# Key listener statics
LISTENER_COMBINATION = [
    keyboard.Key.ctrl_l,
    keyboard.Key.cmd,
    keyboard.KeyCode(char='c')
]

# Server statics
SERVER_ARE_YOU_RUNNING = b'Running?'
SERVER_YES = b'Yes'
SERVER_STOP = b'Stop'
SERVER_LOCK_FILE = os.path.dirname(os.path.realpath(__file__)) + '\.server_lock'

# Setup calls for the MainThread to catch to keep the GUI in the MainThread
openGUIEvent = threading.Event()
openGUIEvent.openGUI = False


def start_listener():
    """ Start an instance of ListenerThread so the main thread can continue """
    listener_thread = ListenerThread()
    listener_thread.start()


def stop_listener():
    """ Tell the current running server to stop, must check is_listener_running() before calling """
    f = open(SERVER_LOCK_FILE, 'r')
    port = int(f.readline())
    f.close()
    server = ('127.0.0.1', port)

    client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    client_socket.sendto(SERVER_STOP, server)


def is_listener_running():
    """ Check if the listener is running, returns a bool"""
    if os.path.isfile(SERVER_LOCK_FILE):
        # If file exists, get the port
        try:
            f = open(SERVER_LOCK_FILE, 'r')
            port = int(f.readline())
            f.close()
            server = ('127.0.0.1', port)
        except PermissionError:
            return False

        # Send packet and then wait for reply
        client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        client_socket.settimeout(1.0)
        client_socket.sendto(SERVER_ARE_YOU_RUNNING, server)
        try:
            message, server = client_socket.recvfrom(1024)
            if message == SERVER_YES:
                return True
        except socket.timeout:
            pass
        except ConnectionResetError:
            pass

    return False


def setup_listener_auto_start():
    """ Creates startup file with the correct command """
    f = open(STARTUP_FILE, 'w')
    f.write('Set oShell = WScript.CreateObject ("WScript.Shell")\n')
    f.write('oShell.run "' + STARTUP_COMMAND + '", 0, True')
    f.close()


def remove_listener_auto_start():
    """ Removes the startup file """
    os.remove(STARTUP_FILE)


def is_listener_auto_start():
    """ Checks to see if the startup file exists """
    return os.path.isfile(STARTUP_FILE)


class ListenerThread(threading.Thread):
    """ A listener thread for a key combination and a server to close the listener on command """

    keys_pressed = set()
    listener = None

    def __init__(self):
        super(ListenerThread, self).__init__()

    def run(self):
        """ When .start() is called, start the server in another thread then run the listener in this thread """
        server_thread = threading.Thread(target=self.server)
        server_thread.start()
        self.listen() # Just run this in the current thread

    def server(self):
        """ Create a local server. Wait for packets coming and and act accordingly """
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        server_socket.bind(('127.0.0.1', 0))

        f = open(SERVER_LOCK_FILE, 'w')
        f.write(str(server_socket.getsockname()[1]))
        f.close()

        while True:
            message, address = server_socket.recvfrom(1024)
            if message == SERVER_ARE_YOU_RUNNING:
                server_socket.sendto(SERVER_YES, address)
            elif message == SERVER_STOP:
                self.listener.stop()
                os.remove(SERVER_LOCK_FILE)
                break

        # Call the GUI event with openGUI = False, will end the loop
        openGUIEvent.openGUI = False
        openGUIEvent.set()

    def listen(self):
        """ Start a listener and join it to this thread """
        with keyboard.Listener(on_press=self.on_press, on_release=self.on_release) as self.listener:
            self.listener.join()

    def on_press(self, key):
        """ When a key is pressed, check if it completes the combination of key presses """
        if key in LISTENER_COMBINATION:
            self.keys_pressed.add(key)
            if all([key in self.keys_pressed for key in LISTENER_COMBINATION]):
                # Make sure there isn't already a GUI_Thread running
                for thread in threading.enumerate():
                    if thread.getName() == 'GUI_Thread' and thread.is_alive():
                        break
                else:
                    # Call GUI in MainThread
                    openGUIEvent.openGUI = True
                    openGUIEvent.set()

    def on_release(self, key):
        """ When a key is released, remove it from the set remembering keys"""
        if key in LISTENER_COMBINATION and key in self.keys_pressed:
            self.keys_pressed.remove(key)


if __name__ == "__main__":
    if not is_listener_running():
        start_listener()



from multi_clipboard.db import DatabaseManager
from multi_clipboard import ui
from multi_clipboard import listener
from multi_clipboard import utils
import argparse


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--set', type=int, action='store', help='set the current clipboard')
    parser.add_argument('-c', '--clear', type=utils.check_clear_arg, action='append', help='clear a clipboard')
    parser.add_argument('--start-listener', help='start the listener', action='store_true')
    parser.add_argument('--stop-listener', help='stop the listener if it is running', action='store_true')
    parser.add_argument('--current', help='get the current clipboard the user is on', action='store_true')
    args = parser.parse_args()

    db_manager = DatabaseManager()

    utils.store_clipboard(db_manager)

    if args.clear is None and args.set is None and not args.start_listener and not args.stop_listener and not args.current:
        # If there were no arguments passed, open the UI
        ui.show_clipboard_selector(db_manager)

    if args.set is not None:
        stored_clipboard_data = db_manager.get_clipboard(args.set)
        if stored_clipboard_data is None:
            # If the requested clipboard id doesn't exist: put the users clipboard in the new clipboard (don't touch old)
            db_manager.current_clipboard = args.set
            utils.store_clipboard(db_manager)
            print('Clipboard switched to ' + str(args.set))
            print('This clipboard did not exist so your clipboard was not changed')
        else:
            # If the requested clipboard id does exist: load the clipboard data
            utils.set_clipboard(db_manager, args.set)
            print('Clipboard switched to ' + str(args.set))
            print('Data from clipboard ' + str(args.set) + ' has been loaded into your clipboard')

    if args.clear is not None and len(args.clear) > 0:
        clipboard_ids = args.clear
        if '*' in args.clear:
            # If * is passed, get all the clipboards to be deleted
            clipboard_ids = db_manager.get_clipboard_ids()

        utils.delete_stored_clipboards(db_manager, clipboard_ids)

        if '*' in args.clear:
            print('All clipboard cleared')
            print('You are now on clipboard 0')
        else:
            print('Clipboards ' + str(args.clear) + 'cleared')
            if db_manager.current_clipboard in args.clear:
                print(
                    'Clipboard ' + str(db_manager.current_clipboard) + ' will be re-created as you are currently on it')

    if args.start_listener:
        # If --start-listener is passed, start the listener
        if not listener.is_listener_running():
            listener.start_listener()
            print('Listener has been started')
        else:
            print('Listener is already running')

    if args.stop_listener:
        # If --stop-listener is passed, stop the listener
        if listener.is_listener_running():
            listener.stop_listener()
            print('Listener has been stopped')
        else:
            print('Listener isn\'t running')

    if args.current:
        # Show the user what clipboard they are currently on
        print('Currently you are on clipboard ' + str(db_manager.current_clipboard))

    # If the listener is running, wait for a call to open the GUI
    while listener.is_listener_running() or args.start_listener:
        args.start_listener = False # Compensate for next loop
        listener.openGUIEvent.wait()
        # When the event is set, check if it was to open the GUI
        if listener.openGUIEvent.openGUI:
            utils.store_clipboard(db_manager)
            ui.show_clipboard_selector(db_manager)
        listener.openGUIEvent.openGUI = False
        listener.openGUIEvent.clear()


if __name__ == '__main__':
    run()



import io
import win32clipboard
from PIL import ImageGrab

CF_IMAGES = win32clipboard.CF_DIBV5
CF_FILES = 49376
CF_HTML = win32clipboard.RegisterClipboardFormat("HTML Format")
CF_UNICODE_TEXT = win32clipboard.CF_UNICODETEXT
CF_RTF = win32clipboard.RegisterClipboardFormat("Rich Text Format")
CF_PLAIN_TEXT = win32clipboard.CF_TEXT

# Store the supported formats in order they must be read in
SUPPORTED_CF = [
    # 49384, # 49384: Focus on publisher items # TODO Find something that won't influence the rest - we already had it?
    CF_IMAGES, # 17: Images
    CF_FILES, # 49376: Files/Folders
    CF_HTML, # 49384: HTML
    CF_UNICODE_TEXT, # 13: Unicode Text
    CF_RTF, # 49285: Rich text
    CF_PLAIN_TEXT # 1: Text
]

# Formats to use for getting previews
CF_PREVIEW_RELATIONS = {
    CF_IMAGES: 0,
    CF_FILES: 15,
    CF_HTML: 13,
    CF_UNICODE_TEXT: 13,
    CF_RTF: 13,
    CF_PLAIN_TEXT: 13
}


def get_clipboard_type():
    """ Get the most recommended clipboard format, return None if we can't handle it """
    win32clipboard.OpenClipboard()
    supported_cf = None
    for _format in SUPPORTED_CF:
        if win32clipboard.IsClipboardFormatAvailable(_format):
            supported_cf = _format
            break
    win32clipboard.CloseClipboard()
    return supported_cf


def get_clipboard():
    """ Return the current data in the clipboard """
    format_available = get_clipboard_type()
    win32clipboard.OpenClipboard()
    contents = win32clipboard.GetClipboardData(format_available)
    win32clipboard.CloseClipboard()
    return contents


def get_clipboard_preview():
    """ Returns a string of the clipboard preview, will be an empty string if no preview format is available """
    clipboard_type = get_clipboard_type()
    preview_type = CF_PREVIEW_RELATIONS[clipboard_type]
    if preview_type == 0:
        if clipboard_type == CF_IMAGES:
            # If the clipboard type is an image, get the bytes using Pillow in JPEG format
            img = ImageGrab.grabclipboard()
            data = io.BytesIO()
            img.save(data, format='JPEG')
            return data.getvalue()
        else:
            return ''
    # Attempt to get preview
    win32clipboard.OpenClipboard()
    try:
        contents = win32clipboard.GetClipboardData(preview_type)
    except:
        contents = "Can't get preview"
    win32clipboard.CloseClipboard()
    return str(contents)


def set_clipboard(_type, _data, preview_extra):
    """ Set the clipboard passing a format (type) and data """
    win32clipboard.OpenClipboard()
    win32clipboard.EmptyClipboard()
    win32clipboard.SetClipboardData(_type, _data)
    # Preview extra is so we can keep reading the preview later
    if CF_PREVIEW_RELATIONS[_type] != 0:
        win32clipboard.SetClipboardData(CF_PREVIEW_RELATIONS[_type], preview_extra)
    win32clipboard.CloseClipboard()



from multi_clipboard import clipboard
from multi_clipboard import ui
import argparse
import sys
from PyQt5.QtWidgets import QFileDialog
import os
from pathlib import Path


def check_clear_arg(value):
    """ Make sure the value passed in with the -c flag is either an integer or star (*) """
    if value == '*':
        return '*'
    else:
        try:
            return int(value)
        except ValueError:
            raise argparse.ArgumentTypeError("%s is not an integer or '*'" % value)


def delete_stored_clipboards(db_manager, clipboard_ids):
    """ Takes a list of clipboards to be deleted, will make sure there is still a clipboard left """

    # Delete all clipboards requested
    for clipboard_id in clipboard_ids:
        db_manager.remove_clipboard(clipboard_id)

    # If there are no more clipboards left, set the current clipboard to 0 and put the users clipboard into it
    if len(db_manager.get_clipboard_ids()) < 1:
        db_manager.current_clipboard = 0

    store_clipboard(db_manager)


def store_clipboard(db_manager):
    """ Stores the users current clipboard into the current clipboard in the database """
    user_clipboard_type = clipboard.get_clipboard_type()
    if user_clipboard_type is None:
        # If we can't support this clipboard, show an error (will stop after dialog closed)
        ui.show_unsupported_clipboard_warning()
        sys.exit()
    user_clipboard_contents = clipboard.get_clipboard()
    user_clipboard_preview = clipboard.get_clipboard_preview()
    current_clipboard_id = db_manager.current_clipboard
    db_manager.set_clipboard(current_clipboard_id, user_clipboard_type, user_clipboard_contents, user_clipboard_preview)


def set_clipboard(db_manager, clipboard_id):
    """ Puts a stored clipboard on the clipboard """
    requested_clipboard = db_manager.get_clipboard(clipboard_id)
    clipboard.set_clipboard(requested_clipboard['type'], requested_clipboard['content'], requested_clipboard['preview'])
    db_manager.current_clipboard = clipboard_id


def create_blank_clipboard(db_manager):
    """ Creates a new clipboard with an id of the current maximum id """
    _id = db_manager.get_next_clipboard_value()
    db_manager.set_clipboard(_id, clipboard.CF_PLAIN_TEXT, '', '')


def create_shortcut(parent=None):
    desktop = os.path.expanduser("~/Desktop")
    options = QFileDialog.Options()
    filename, _filter = QFileDialog.getSaveFileName(
        parent,
        "Select Directory",
        desktop,
        "Visual Basic (*.vbs)",
        options=options
    )
    if filename != '':
        project_directory = str(Path(__file__).parent.parent)
        f = open(filename, 'w')
        f.write('Set oShell = WScript.CreateObject ("WScript.Shell")\n')
        f.write('oShell.CurrentDirectory = "' + project_directory + '"\n')
        f.write('oShell.run "' + sys.executable + ' -m multi_clipboard", 0, True\n')
        f.close()



import sqlite3
import os


DATABASE_LOCATION = os.path.dirname(os.path.realpath(__file__)) + '\\clipboards.sqlite'


class DatabaseManager:
    """ Manages clipboard storage and settings for the GUI """

    def __init__(self):
        self._open_connection()
        self.cursor.execute('CREATE TABLE IF NOT EXISTS clipboards (id INTEGER PRIMARY KEY, type INTEGER, content BLOB, preview BLOB);')
        self.cursor.execute('CREATE TABLE IF NOT EXISTS settings (id INTEGER PRIMARY KEY, current_clipboard INTEGER, close_on_select INTEGER, stay_on_top INTEGER, opacity INTEGER, disable_frame INTEGER);')
        self.cursor.execute('SELECT COUNT(*) FROM settings')
        if self.cursor.fetchone()[0] < 1:
            self.cursor.execute('INSERT INTO settings VALUES (0, 0, 1, 1, 85, 1)')
        self._close_connection()

    def _open_connection(self):
        """ Open a connection to the database """
        self.connection = sqlite3.connect(DATABASE_LOCATION)
        self.connection.row_factory = sqlite3.Row
        self.cursor = self.connection.cursor()

    def _close_connection(self):
        """ Commit and close the current connection to the database"""
        self.connection.commit()
        self.connection.close()

    # Settings

    def _get_setting(self, setting):
        """ Get a particular setting value from the settings table """
        self._open_connection()
        self.cursor.execute('SELECT ' + setting + ' FROM settings')
        row = self.cursor.fetchone()
        self._close_connection()
        return row[0]

    def _set_setting(self, setting, value):
        """ Set a particular settings value in the settings table """
        self._open_connection()
        self.cursor.execute('UPDATE settings SET ' + setting + ' = ? WHERE id = 0', (value, ))
        self._close_connection()

    @property
    def current_clipboard(self):
        return self._get_setting('current_clipboard')

    @current_clipboard.setter
    def current_clipboard(self, value):
        self._set_setting('current_clipboard', value)

    @property
    def close_on_select(self, ):
        return bool(self._get_setting('close_on_select'))

    @close_on_select.setter
    def close_on_select(self, value):
        self. _set_setting('close_on_select', int(value))

    @property
    def stay_on_top(self):
        return bool(self._get_setting('stay_on_top'))

    @stay_on_top.setter
    def stay_on_top(self, value):
        self._set_setting('stay_on_top', int(value))

    @property
    def opacity(self):
        return self._get_setting('opacity') / 100

    @opacity.setter
    def opacity(self, value):
        self._set_setting('opacity', value * 100)

    @property
    def disable_frame(self):
        return bool(self._get_setting('disable_frame'))

    @disable_frame.setter
    def disable_frame(self, value):
        self._set_setting('disable_frame', int(value))

    # Clipboards

    def get_next_clipboard_value(self):
        """ Find the next value we can use for a new clipboard """
        self._open_connection()
        self.cursor.execute('SELECT MAX(id) + 1 FROM clipboards')
        next_id = self.cursor.fetchone()[0]
        self._close_connection()
        return next_id

    def get_clipboard(self, clipboard_id):
        """ Get the data associated with a clipboard id """
        self._open_connection()
        self.cursor.execute('SELECT type, content, preview FROM clipboards WHERE id = ?', (clipboard_id,))
        data = self.cursor.fetchone()
        self._close_connection()
        if data is None:
            return None
        return {
            'type' : data['type'],
            'content' : data['content'],
            'preview' : data['preview']
        }

    def set_clipboard(self, clipboard_id, clipboard_type, clipboard_content, preview=''):
        """ Set the data of a clipboard id """
        self._open_connection()
        self.cursor.execute('INSERT OR REPLACE INTO clipboards VALUES (?, ?, ?, ?)', (clipboard_id, clipboard_type, clipboard_content, preview))
        self._close_connection()

    def remove_clipboard(self, clipboard_id):
        """ Delete a clipboard id from the clipboards table """
        self._open_connection()
        self.cursor.execute('DELETE FROM clipboards WHERE id = ?', (clipboard_id,))
        self._close_connection()

    def get_clipboard_ids(self):
        """ Return a list of all the clipboard ids from the clipboards table """
        self._open_connection()
        self.cursor.execute('SELECT id FROM clipboards')
        rows = self.cursor.fetchall()
        self._close_connection()
        return [row['id'] for row in rows]



from io import open
from setuptools import setup

setup(
    name='multi-clipboard',
    version='1.3.0',
    url='https://github.com/brentvollebregt/mutli-clipboard',
    license='MIT',
    author='Brent Vollebregt',
    author_email='brent@nitratine.net',
    description='Switch clipboard contents using a simple GUI',
    long_description=''.join(open('README.md', encoding='utf-8').readlines()),
    long_description_content_type='text/markdown',
    keywords=['gui', 'clipboard'],
    packages=['multi_clipboard'],
    include_package_data=True,
    install_requires=['pywin32', 'PyQt5', 'Pillow', 'pynput'],
    python_requires='>=3.5',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: Microsoft :: Windows',
    ],
    entry_points={
        'console_scripts': [
            'multi-clipboard=multi_clipboard.__main__:run',
        ],
    },
)



from multi_clipboard import clipboard
from multi_clipboard import utils
from multi_clipboard import listener
from PyQt5 import QtCore, QtGui, QtWidgets
import sys
import os


# Location of the images folder
IMAGES_FOLDER = os.path.dirname(os.path.realpath(__file__)) + '\\images'

# Static values for the GUI - for customisation
GRID_SPACING = 6
CLIPBOARD_LABEL_SIZE = 130
BUTTONS_SPACING = 4
BUTTON_IMAGE_SIZE_REDUCTION = 15
SETTINGS_DISTANCE_ABOVE_PARENT = 5


class ClipboardSelector(QtWidgets.QWidget):
    """ The main window for the clipboard selection GUI"""

    focus_enabled = False
    current_focus = None

    def __init__(self, db_mgr):
        super(ClipboardSelector, self).__init__()
        self.db_manager = db_mgr

        # Setup window: title, flags, stylesheet, opacity
        self.setWindowTitle('Multi Clipboard')
        if self.db_manager.stay_on_top and self.db_manager.disable_frame:
            self.setWindowFlags(QtCore.Qt.FramelessWindowHint | QtCore.Qt.WindowStaysOnTopHint)
        elif self.db_manager.stay_on_top:
            self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        elif self.db_manager.disable_frame:
            self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        self.setStyleSheet(
            'QMenu { border: 1px solid #000; }'
            'QMenu::item { padding: 2px 20px 2px 20px; }'
            'QMenu::item:selected { color: #000000; }'
            'QWidget { color: #b1b1b1; background-color: #323232; }'
            'QWidget:item:selected { background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a); }'
            'QLabel { border: 1px solid #b1b1b1; }'
            'QLabel:hover { border: 2px solid #ffaa00; }'
            'QLabel:focus { border: 2px solid #ffaa00; }'
        )
        self.setWindowOpacity(self.db_manager.opacity)

        clipboard_ids = self.db_manager.get_clipboard_ids()

        # Calculating grid and window size
        clipboards_total = len(clipboard_ids)
        if clipboards_total < 6:
            rows = 1
            cols = clipboards_total + 1
        else:
            rows = int(clipboards_total / 6) + 1
            cols = 6
        self.setFixedSize(
            ((CLIPBOARD_LABEL_SIZE + GRID_SPACING) * cols) + GRID_SPACING,
            ((CLIPBOARD_LABEL_SIZE + GRID_SPACING) * rows) + GRID_SPACING
        )

        # Setup grid layout
        self.grid_layout = QtWidgets.QGridLayout()
        self.setLayout(self.grid_layout)
        self.grid_layout.setSpacing(GRID_SPACING)
        self.grid_layout.setContentsMargins(GRID_SPACING, GRID_SPACING, GRID_SPACING, GRID_SPACING)

        # Add clipboards
        for position, _id in enumerate(clipboard_ids):
            if position < 5:
                # If there are less than 5 clipboards (to account for settings button)
                _col = position
                _row = 0
            else:
                _row = int((position + 1) / 6)
                _col = int((position + 1) % 6)

            self.grid_layout.addWidget(self.create_clipboard_label(_id), _row, _col)

        # Add buttons
        if clipboards_total < 6:
            self.grid_layout.addLayout(self.create_buttons(), 0, clipboards_total)
        else:
            self.grid_layout.addLayout(self.create_buttons(), 0, 5)

        self.centre()
        self.show()

    def create_clipboard_label(self, clipboard_id):
        """ Create a label for a particular clipboard id """
        label = QtWidgets.QLabel()
        label.setFixedSize(CLIPBOARD_LABEL_SIZE, CLIPBOARD_LABEL_SIZE)
        label.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        label.clipboard_id = clipboard_id

        # Get the clipboard and check if it is valid
        clipboard_contents = self.db_manager.get_clipboard(clipboard_id)
        if clipboard_contents is None:
            clipboard_contents = {'type' : -1, 'content' : '', 'preview' : 'No Preview'}

        # Decide on what is being shown in the clipboard (preview)
        if clipboard.CF_PREVIEW_RELATIONS[clipboard_contents['type']] != 0:
            label.setText(clipboard_contents['preview'])
            label.setMargin(3)
        elif clipboard_contents['type'] == clipboard.CF_IMAGES:
            image = QtGui.QPixmap()
            image.loadFromData(clipboard_contents['preview'], 'JPEG')
            image = image.scaled(CLIPBOARD_LABEL_SIZE, CLIPBOARD_LABEL_SIZE, QtCore.Qt.KeepAspectRatio)
            label.setPixmap(image)
        else:
            label.setText('Preview not available')
            label.setMargin(3)

        # Formatting
        label.setAlignment(QtCore.Qt.AlignCenter)
        label.setWordWrap(True)
        label.setAlignment(QtCore.Qt.AlignCenter)
        label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        label.setToolTip("Clipboard: " + str(clipboard_id))

        # If this is the currently selected clipboard, show the user
        if self.db_manager.current_clipboard == clipboard_id:
            label.setStyleSheet(
                'QLabel {border: 1px solid #ffaa00;}'
                'QLabel:hover {border: 2px solid #ffaa00;}'
                'QLabel:focus {border: 2px solid #ffaa00;}'
            )

        # On click
        label_click = self.LabelClick(clipboard_id, self, label)
        label.mousePressEvent = label_click.on_click

        return label

    def create_buttons(self):
        """ Create the buttons for the right hand side of the GUI """
        # Create base label
        layout = QtWidgets.QGridLayout()
        layout.setGeometry(QtCore.QRect(0, 0, CLIPBOARD_LABEL_SIZE, CLIPBOARD_LABEL_SIZE))
        layout.setSpacing(BUTTONS_SPACING)
        layout.setContentsMargins(0, 0, 0, 0)

        # Size of each label with spacing considered
        label_size = (CLIPBOARD_LABEL_SIZE - BUTTONS_SPACING) / 2

        # Data to setup buttons to stop repeating lines [img, onclick, row, col, tooltip]
        button_data = [
            [IMAGES_FOLDER + '\\add.png', self.add_button, 0, 0, 'Add a new empty clipboard'],
            [IMAGES_FOLDER + '\\settings.png', self.settings_button, 0, 1, 'Open the settings'],
            [IMAGES_FOLDER + '\\delete.png', self.clear_button, 1, 0, 'Delete all clipboards'],
            [IMAGES_FOLDER + '\\close.png', self.close_button, 1, 1, 'Close this window']
        ]

        # Setup each button with data provided above
        for button in button_data:
            tmp_btn = QtWidgets.QLabel()
            tmp_btn.setFixedSize(label_size, label_size)
            tmp_btn.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
            icon = QtGui.QPixmap(button[0])
            icon = icon.scaled(label_size - BUTTON_IMAGE_SIZE_REDUCTION, label_size - BUTTON_IMAGE_SIZE_REDUCTION, QtCore.Qt.KeepAspectRatio)
            tmp_btn.setPixmap(icon)
            tmp_btn.setAlignment(QtCore.Qt.AlignCenter)
            tmp_btn.setToolTip(button[4])
            tmp_btn.mousePressEvent = button[1]
            layout.addWidget(tmp_btn, button[2], button[3])

        return layout

    def centre(self):
        """ Centre the window relative to the active display """
        # Thanks to https://stackoverflow.com/questions/20243637/pyqt4-center-window-on-active-screen
        frame_gm = self.frameGeometry()
        screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())
        center_point = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
        frame_gm.moveCenter(center_point)
        self.move(frame_gm.topLeft())

    def refresh(self):
        """ Refresh the whole GUI"""
        self.close()
        self.__init__(self.db_manager)

    def add_button(self, event):
        """ Event for the add button """
        utils.create_blank_clipboard(self.db_manager)
        self.refresh()

    def close_button(self, event):
        """ Event for the close button """
        if event.button() == 1:
            self.close()

    def clear_button(self, event):
        """ Event for the clear button """
        utils.delete_stored_clipboards(self.db_manager, self.db_manager.get_clipboard_ids())
        self.refresh()

    def settings_button(self, event):
        """ Event for the settings button """
        settings_window = ClipboardSelector.SettingsWindow(self)
        settings_window.show()

    def keyPressEvent(self, event):
        """ Detecting keypress events globally to setup tab selection"""
        # If tab has been pressed and the tab selections havent been setup
        if event.key() == QtCore.Qt.Key_Tab and not self.focus_enabled:
            self.focus_enabled = True
            child_count = self.grid_layout.count()
            # Set all clipboards focus policies
            for child_index in range(child_count):
                if self.grid_layout.itemAt(child_index).widget() is not None:
                    self.grid_layout.itemAt(child_index).widget().setFocusPolicy(QtCore.Qt.StrongFocus)
            # Set the focus to the first clipboard
            self.grid_layout.itemAt(0).widget().setFocus()

        # If enter is pressed and we have setup tab selection
        elif event.key() in [QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return] and self.focus_enabled:
            # Send a left click event to the currently focused widget
            event = QtCore.QEvent(QtCore.QEvent.MouseButtonPress)
            event.button = lambda: 1
            self.focusWidget().mousePressEvent(event)

    class LabelClick:
        """ Handler for left and right clicks on a label """

        def __init__(self, clipboard_id, parent, label):
            self.clipboard_id = clipboard_id
            self.parent = parent
            self.label = label

        def on_click(self, event):
            """ When clicked, decide what to do """
            if event.button() == 1:
                # If this is a left click, set the clipboard and close if setting is true
                utils.set_clipboard(self.parent.db_manager, self.clipboard_id)
                if self.parent.db_manager.close_on_select:
                    self.parent.close()
                else:
                    self.parent.refresh()
            elif event.button() == 2:
                # If this is a right click, open context menu
                menu = QtWidgets.QMenu(self.label)
                remove_action = QtWidgets.QAction('remove', self.label)
                menu.addAction(remove_action)
                menu.addSeparator()
                switch_action = QtWidgets.QAction('switch', self.label)
                menu.addAction(switch_action)
                action = menu.exec_(self.label.mapToGlobal(event.pos()))
                if action is not None:
                    if action == remove_action:
                        utils.delete_stored_clipboards(self.parent.db_manager, [self.clipboard_id])
                        self.parent.refresh()
                    elif action == switch_action:
                        utils.set_clipboard(self.parent.db_manager, self.clipboard_id)
                        self.parent.refresh()

    class SettingsWindow(QtWidgets.QWidget):
        """ Settings window for the main clipboard GUI """

        SETTINGS_GRID_SPACING = 4
        SETTINGS_TILE_SIZE = 60

        def __init__(self, parent):
            super(ClipboardSelector.SettingsWindow, self).__init__()
            self.parent = parent

            # Setup settings window
            self.setWindowFlags(QtCore.Qt.FramelessWindowHint | QtCore.Qt.WindowStaysOnTopHint)
            self.setStyleSheet('QWidget {color: #b1b1b1; background-color: #323232; border: 0px;}')
            self.setWindowOpacity(1)

            # Setup grid layout
            self.grid_layout = QtWidgets.QGridLayout()
            self.setLayout(self.grid_layout)
            self.grid_layout.setSpacing(self.SETTINGS_GRID_SPACING)
            self.grid_layout.setContentsMargins(
                self.SETTINGS_GRID_SPACING,
                self.SETTINGS_GRID_SPACING,
                self.SETTINGS_GRID_SPACING,
                self.SETTINGS_GRID_SPACING
            )

            # Calculate window size based off **static** value of how many buttons/inputs we will have
            items = 8
            self.setFixedSize(
                (items * (self.SETTINGS_GRID_SPACING + self.SETTINGS_TILE_SIZE)) + self.SETTINGS_GRID_SPACING,
                self.SETTINGS_TILE_SIZE + (2 * self.SETTINGS_GRID_SPACING)
            )

            # Setup general toggle buttons
            self.close_on_select_button = self.create_basic_button(
                'Close on Select',
                self.close_on_select_button_click,
                'Close the widow when a clipboard is selected'
            )
            self.grid_layout.addWidget(self.close_on_select_button, 0, 0)

            self.stay_on_top_button = self.create_basic_button(
                'Stay on Top',
                self.stay_on_top_button_click,
                'Keep the window on top of all windows'
            )
            self.grid_layout.addWidget(self.stay_on_top_button, 0, 1)

            self.disable_frame_button = self.create_basic_button(
                'Disable Frame',
                self.disable_frame_button_click,
                'Disable the frame around the window\n(includes the minimise and close button)'
            )
            self.grid_layout.addWidget(self.disable_frame_button, 0, 2)

            # Setup opacity QSpinBox
            self.opacity_spin = QtWidgets.QSpinBox()
            self.opacity_spin.setFixedSize(self.SETTINGS_TILE_SIZE, self.SETTINGS_TILE_SIZE)
            self.opacity_spin.setMaximum(100)
            self.opacity_spin.setMinimum(0)
            self.opacity_spin.setStyleSheet('QSpinBox {border: 1px solid #ffffff; font: 16pt; color: white;}')
            self.opacity_spin.setToolTip('Set the opacity of the main window')
            self.opacity_spin.valueChanged.connect(self.opacity_edit)
            self.grid_layout.addWidget(self.opacity_spin, 0, 3)

            # Setup listener toggle buttons
            self.toggle_listener_button = self.create_basic_button(
                'Toggle Listener',
                self.toggle_listener_button_click,
                'Turn the listener on/off\nPress Ctrl + Windows + C to open multi clipboard when on'
            )
            self.grid_layout.addWidget(self.toggle_listener_button, 0, 4)

            self.toggle_listener_auto_start_button = self.create_basic_button(
                'Listener Autostart',
                self.toggle_listener_auto_start_button_click,
                'Put a file in the startup folder to start the listener on startup'
            )
            self.grid_layout.addWidget(self.toggle_listener_auto_start_button, 0, 5)

            # Setup Create Shortcut button
            self.create_shortcut_button = self.create_basic_button(
                'Create Shortcut',
                self.create_shortcut,
                'Create a VB Script that can be double clicked to run this project'
            )
            self.grid_layout.addWidget(self.create_shortcut_button, 0, 6)
            self.create_shortcut_button.setStyleSheet(
                'QLabel {border: 1px solid #ffffff; font: 8pt; color: white;}'
                'QLabel:hover {border: 2px solid #ffaa00; color: #ffaa00;}'
            )

            # Setup close button
            icon = QtGui.QPixmap(IMAGES_FOLDER + '\\close.png')
            icon = icon.scaled(self.SETTINGS_TILE_SIZE, self.SETTINGS_TILE_SIZE, QtCore.Qt.KeepAspectRatio)
            self.close_label = QtWidgets.QLabel()
            self.close_label.setFixedSize(self.SETTINGS_TILE_SIZE, self.SETTINGS_TILE_SIZE)
            self.close_label.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
            self.close_label.setStyleSheet(
                'QLabel {border: 1px solid #ffffff; font: 8pt; color: white;}'
                'QLabel:hover {border: 2px solid #ffaa00; color: #ffaa00;}'
            )
            self.close_label.setPixmap(icon)
            self.close_label.setAlignment(QtCore.Qt.AlignCenter)
            self.close_label.setToolTip('Close settings')
            self.close_label.mousePressEvent = self.close_button_click
            self.grid_layout.addWidget(self.close_label, 0, 7)

            self.position_above_parent()
            self.set_values()

        def create_basic_button(self, text, onclick, tooltip):
            """ Create a basic settings button with text and a onclick method """
            tmp_btn = QtWidgets.QLabel()
            tmp_btn.setFixedSize(self.SETTINGS_TILE_SIZE, self.SETTINGS_TILE_SIZE)
            tmp_btn.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
            tmp_btn.setText(text)
            tmp_btn.setWordWrap(True)
            tmp_btn.setMargin(3)
            tmp_btn.setAlignment(QtCore.Qt.AlignCenter)
            tmp_btn.mousePressEvent = onclick
            tmp_btn.setToolTip(tooltip)
            return tmp_btn

        def position_above_parent(self):
            """ Position the settings GUI above the parent (main clipboard GUI) """
            parents_geometry = self.parent.frameGeometry()
            parent_top = parents_geometry.y()
            parent_center = parents_geometry.x() + (parents_geometry.width() / 2)

            # Calculate positions of new window based off parent
            my_geometry = self.frameGeometry()
            my_x = parent_center - (my_geometry.width() / 2)
            my_y = parent_top - my_geometry.height() - SETTINGS_DISTANCE_ABOVE_PARENT
            self.move(my_x, my_y)

        def set_values(self):
            """ Update the display/value for each label depending on real states """
            self.opacity_spin.setValue(self.parent.db_manager.opacity * 100)

            settings_label_pairs = [
                [self.parent.db_manager.close_on_select, self.close_on_select_button],
                [self.parent.db_manager.stay_on_top, self.stay_on_top_button],
                [self.parent.db_manager.disable_frame, self.disable_frame_button],
                [listener.is_listener_running(), self.toggle_listener_button],
                [listener.is_listener_auto_start(), self.toggle_listener_auto_start_button],
            ]

            # Go through each index of settings_label_pairs checking value and then setting label style sheet
            for pair in settings_label_pairs:
                if pair[0]:
                    pair[1].setStyleSheet(
                        'QLabel {border: 1px solid #ffaa00; font: 8pt; color: #ffaa00;}'
                        'QLabel:hover {border: 2px solid #ffaa00; color: #ffaa00;}'
                    )
                else:
                    pair[1].setStyleSheet(
                        'QLabel {border: 1px solid #ffffff; font: 8pt; color: white;}'
                        'QLabel:hover {border: 2px solid #ffaa00; color: #ffaa00;}'
                    )

        def close_on_select_button_click(self, e):
            """ Event handler for close_on_select_button """
            self.parent.db_manager.close_on_select = not self.parent.db_manager.close_on_select
            self.set_values()

        def stay_on_top_button_click(self, e):
            """ Event handler for stay_on_top_button """
            self.parent.db_manager.stay_on_top = not self.parent.db_manager.stay_on_top
            self.set_values()

        def disable_frame_button_click(self, e):
            """ Event handler for disable_frame_button """
            self.parent.db_manager.disable_frame = not self.parent.db_manager.disable_frame
            self.set_values()

        def opacity_edit(self, e):
            """ Event handler for the edit of opacity_edit """
            self.parent.db_manager.opacity = self.opacity_spin.value() / 100
            self.set_values()

        def toggle_listener_button_click(self, e):
            """ Event handler for toggle_listener_button """
            if listener.is_listener_running():
                listener.stop_listener()
            else:
                listener.start_listener()
            self.set_values()

        def toggle_listener_auto_start_button_click(self, e):
            """ Event handler for toggle_listener_auto_start_button """
            if listener.is_listener_auto_start():
                listener.remove_listener_auto_start()
            else:
                listener.setup_listener_auto_start()
            self.set_values()

        def create_shortcut(self, e):
            utils.create_shortcut(self)

        def close_button_click(self, e):
            """ Close the window when the close button is pressed """
            if e.button() == 1:
                self.close()


def show_clipboard_selector(db_manager):
    """ Open the main GUI application """
    app = QtWidgets.QApplication(sys.argv)
    cs = ClipboardSelector(db_manager)
    app.exec_()


def show_unsupported_clipboard_warning():
    """ Show a message warning the user about an unsupported clipboard """
    app = QtWidgets.QApplication([])
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Critical)
    msg.setText(
        'The current item on your clipboard is not supported.\n'
        'The application is unable to detect the type of the clipboard so it must stop.\n\n'
        'No data has been lost. Changing the contents of the clipboard will fix this'
    )
    msg.setWindowTitle("Fatal Error")
    msg.show()
    msg.raise_()
    msg.activateWindow()
    app.exec_()
